// <pre>

import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.*;
import render.*;

class ParsedString {
	public String str;
	public int numtokens;
}

public class SLAPlayer extends RenderApplet {

	Vector pollys, colors, buttons, commands, lines;
	final int cmdBufferSize = 10;
	int selected;
	Hashtable defines;
	StreamTokenizer st;
	double scriptTime;
	double camtheta = 0. , camphi = 0;
	
	final int TT_EOF = StreamTokenizer.TT_EOF;
	final int TT_EOL = StreamTokenizer.TT_EOL;
	final int TT_NUMBER = StreamTokenizer.TT_NUMBER;
	final int TT_WORD = StreamTokenizer.TT_WORD;

	final Color buttonColor = Color.yellow;
	final Color textColor = Color.black;
	final Color highlightColor = Color.white;
	final int buttonWidth = 60;
	final int buttonHeight = 20;

	// DEBUG
/*	public static void main(String args[]) {
		SLAPlayer s = new SLAPlayer();
		s.initialize();
	}
*/
	public void initialize() {

		if (st == null) {
			URL source;
			InputStreamReader isr;
			String s = "";
			int c;
			try {
				source = new URL(getCodeBase(), "script.slap");
			} catch (MalformedURLException e) { return; }
			try {		
				isr = new InputStreamReader(new DataInputStream(source.openStream())); 
				while ((c = isr.read()) != -1)
					s = s + (char)c;
			} catch (IOException e) { return; }
	
			recordString(s);
			st = new StreamTokenizer(new StringReader(s));

		}
			/*try {
				st = new StreamTokenizer(new FileReader("script.slap"));
			} catch (FileNotFoundException e) {
				System.out.println("Could not open script file : "+e.toString());
				return;
			}*/

		// Initialize stream tokenizer for our language
		st.commentChar('#');     // this is our comment
		st.ordinaryChar('/');    
		st.ordinaryChar('-');    
		st.ordinaryChar('+');    
		st.ordinaryChar('*');    
		st.parseNumbers();
		st.eolIsSignificant(true);  // we _do_ care about lines
		// Case insensitive
		st.lowerCaseMode(true);
		st.slashSlashComments(false);  // c++ style comments
		/* do we want this? */
		st.slashStarComments(true);    // c style comments
		//System.out.println("st was set up");

		defines = new Hashtable();
		pollys = new Vector();
		colors = new Vector();
		buttons = new Vector();
		commands = new Vector();
		selected = -1;

		scriptTime = 0;
	
		addLight(1,1,1, 1,1,1);
		addLight(0,-1,0, 1,1,1);
		setBgColor(.2,.2,.8);            // BACKGROUND COLOR
		setFL(12);                       // CAMERA FOCAL LENGTH
		push();
			translate(-.5,0,-5);
			rotateX(Math.PI/6);
			transform(world);             // INITIAL VIEW ANGLE
		pop();

		world.add().setMaterial(
         (new Material()).setColor(0,1,0).setTransparency(.5));
      	world.child[0].mesh(10, 10).setDoubleSided(true);
      	push();                          // MAKE THE TRANSPARENT FLOOR
			 scale(6,1,4);
			 rotateX(-Math.PI/2);
			 transform(world.child[0]);
		pop();

		performActions(0);   // perform initialization actions	
			
		// clean up before we start
		System.runFinalization();
		System.gc();

	}

	private void clearWorld() {
		int objects = world.child.length;
		for (int i = 0; i < objects; i++) {
			/* delete the first object, they all move down the list */
			world.delete(0);
		}
		pollys.clear();
		colors.clear();
		buttons.clear();
		commands.clear();
		defines.clear();
		/* reset all pollys for their collision avoidance */
		Polly.clearPollys();
		Polly.clearWalls();
	}
			


	private void performActions(double time) {
		try {
			while(st.ttype != TT_EOF && scriptTime <= time) {
				if (st.nextToken() == TT_EOL) { /* empty line */
					continue;
				}
				if (st.ttype == TT_NUMBER) {
					System.out.println("Not expecting number at line "+
						st.lineno());
					System.out.println("Aborting processing script.");
					return;
				}
				if (st.ttype == TT_WORD) {
					if (st.sval.equals("define")) {
						if (!readDefine(st))
							continue;
					}
					else if (st.sval.equals("time")) {
						if (!readTime(st, false))
							continue;
					}
					else if (st.sval.equals("timep")) {
						if (!readTime(st, true))
							continue;
					}
					else if (st.sval.equals("actor")) {
						if (!readActor(st))
							continue;
					}
					else if (st.sval.equals("camera")) {
						if (!readCamera(st))
							continue;
					}
					else {
						System.out.println("Unexpected token: "+st.sval +
							"at line "+st.lineno());
						System.out.println("Aborting processing script.");
						return;
					}
					continue;
				}
				if (st.ttype != TT_EOF) { /* unknown character */
					System.out.println("Unexpected token "+(char)st.ttype+
						"at line "+st.lineno());
					System.out.println("Aborting processing script.");
					return;
				}
			}  // end loop through script
		} catch (IOException e) {}
	}

	// This reads the stream till the end of line, and tries
	// to resolve any symbols according to previously defined keywords
	// Returns the resulting (resolved) string and the number of tokens in it.
	private ParsedString resolveDefines(StreamTokenizer st) throws IOException {
		int type;
		ParsedString ret = new ParsedString();

		// read until the end of the line (or file)
		String value = new String("");
		boolean eol = false;
		int numtokens = 0;
		int lastType = TT_EOL;
		while(!eol) {
			type = st.nextToken();
			numtokens++;

			switch(type) {
				case TT_EOF:
				case TT_EOL:
					eol = true;
					numtokens--;   // it was not a token
					break;

				case TT_WORD:
					// do we have this word in the defines?
					if (defines.containsKey(st.sval)) {
						if (lastType == TT_NUMBER)
							value = value.concat(" ");
						value = value.concat(defines.get(st.sval).toString());
						lastType = TT_NUMBER;
						break;
					}
					// always separate a word out by spaces
					value = value.concat(" "+st.sval+" ");
					lastType = type;
					break;

				case TT_NUMBER:
					// don't paste 2 numbers next to each other
					if (lastType == TT_NUMBER)
						value = value.concat(" "+st.nval);
					else
						value = value.concat(""+st.nval);
					lastType = type;
					break;

				default:   // some character. append it
					value = value.concat(String.valueOf((char)type));
					lastType = type;
					break;
			}
		}

		// DEBUG
//		System.out.println("String is \""+value+"\", with "+numtokens+" keywords");

		// if nothing there, or only one token was read -- there's no way
		// to evaluate it further. Just return it.
		if (numtokens <= 1) {
			ret.str = new String(value);
			ret.numtokens = numtokens;
			return ret;
		}

		// At this point, we should have resolved anything we could
		// resolve, and have a string. We try to arithmetically eval it.
		// First, divide it into substrings :
		int from=0, to;
		value = value.trim();   // trim whitespace
		ret.str = new String("");
		ret.numtokens = 0;
		while (from < value.length()) {
			String str;

			to = value.indexOf(" ", from);
			if (to == -1)
				to = value.length();

			str = value.substring(from, to);
			
			try {
				double result = ArithEval.eval(str);
				ret.str = ret.str.concat(String.valueOf(result)+" ");
				ret.numtokens++;
			}
			catch (IOException e) {   // its not evaluatable
				// tack on the unconverted part
				ret.str = ret.str.concat(value.substring(from, to)+" ");
				// unconverted part has got to be at least 2 tokens
				ret.numtokens += 2;
			}
			from = to+1;
		}
		return ret;
	}

	private boolean readDefine(StreamTokenizer st) throws IOException {
		if (st.nextToken() != TT_WORD) {
			System.out.println("Unexpected define keyword "+st.toString());
			return false;
		}
		// save the keyword
		String keyword = new String(st.sval);
	
		// DEBUG
		//System.out.println("Defining keyword "+keyword);

		// read till the end of the line and resolve any definitions, and
		// perform any arithmetic
		ParsedString value = resolveDefines(st);

		if (value.numtokens == 1) {	
			// See if its fully resolved 
			Double result;
			try {
				result = Double.valueOf(value.str);
			} catch (NumberFormatException e) {
				System.out.println("Unexpected argument to define. String \"" +
					value.str + "\" (after evaluation) is not a number");
				return false;
			}
			defines.put((Object)keyword, (Object)result); 

			// DEBUG
			//System.out.println(keyword + " defined to "+result);
			return true;
		}
		else {
			// If there's more than one thing here, we'll just have
			// to define it as a string
			defines.put((Object)keyword, (Object)value.str);

			//DEBUG
			//System.out.println(keyword + " defined to string '"+value.str+"'");
			return true;
		}
	}

	private boolean readTime(StreamTokenizer st, boolean relative) 
		throws IOException {

		ParsedString value = resolveDefines(st);
		if (value.numtokens != 1) {
			System.out.println("Unexpected argument to time: '"+value.str+" at line "+st.lineno());
			return false;
		}

		// Lets try to get this token 

		Double result;
		try {
			result = Double.valueOf(value.str);
		} catch (NumberFormatException e) {
			System.out.println("Cannot parse argument to time: '"+value.str+" at line "+st.lineno());
			return false;
		}

		if (relative)  // time being set relative to last
			scriptTime += result.doubleValue();
		else {  // absolute time being set
			if (result.doubleValue() < scriptTime) 
				System.out.println("Warning: Time goes backwards at line "+
					st.lineno());
			scriptTime = result.doubleValue();
		}
		return true;
	}

	private boolean readActor(StreamTokenizer st) throws IOException {
		// read the rest of the line to resolve symbols
		ParsedString value = resolveDefines(st);

		StreamTokenizer vst = new StreamTokenizer(new StringReader(value.str));

		// get the number of the actor
		if (vst.nextToken() != TT_NUMBER) {
			System.out.println("Expecting number of actor at line "+
				st.lineno()+" but got "+value.str);
			return false;
		}
		int actor = (int)vst.nval;

		// make sure this actor exists -- make it if not
		if (actor >= pollys.size()) {
			pollys.setSize(actor+1);
			colors.setSize(actor+1);
			buttons.setSize(actor+1);
			commands.setSize(actor+1);
		}
		if (pollys.get(actor) == null) {
			Polly t;
			pollys.add(actor, t=new Polly());
			colors.add(actor, Color.red);    // default is red. we hope.
			buttons.add(actor, new Rectangle(actor * buttonWidth, 
				H - buttonHeight, buttonWidth, buttonHeight));
			commands.add(actor, new Vector());
			world.add(t);
		}

		// get the next keyword
		if (vst.nextToken() != TT_WORD) {
			System.out.println("Unexpected token for actor "+actor+" at line"+
				st.lineno());
			return false;
		}

		Vector curcmd = (Vector)commands.get(actor);
		// remove old commands
		if (curcmd.size() > cmdBufferSize)
			curcmd.remove(cmdBufferSize);
		// record this linenumber
		curcmd.add(0, new Integer(st.lineno()-1));

		if (vst.sval.equals("action")) {
			// action takes one integer -- index into the action array of Polly
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for action at line "+
					st.lineno());
				return false;
			}
		
			((Polly)pollys.get(actor)).action((int)vst.nval);
			return true;
		}

		if (vst.sval.equals("direction")) {
			// direction takes one double -- angle to turn
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for direction at line "+
					st.lineno());
				return false;
			}
			
			((Polly)pollys.get(actor)).setDirection(vst.nval);
			return true;
		}

		if (vst.sval.equals("position")) {
			// position takes two double arguments
			double posx, posy;

			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for position at line "+
					st.lineno());
				return false;
			}
			posx = vst.nval;
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for position at line "+
					st.lineno());
				return false;
			}
			posy = vst.nval;
			
			((Polly)pollys.get(actor)).setPosition(posx, posy);
			return true;
		}

		if (vst.sval.equals("color")) {
			// color takes three double arguments for RGB
			double r, g, b;
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for color at line "+
					st.lineno());
				return false;
			}
			r = vst.nval;
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for color at line "+
					st.lineno());
				return false;
			}
			g = vst.nval;
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for color at line "+
					st.lineno());
				return false;
			}
			b = vst.nval;

			colors.remove(actor);
			colors.add(actor, new Color((float)r,(float)g,(float)b));
			((Polly)pollys.get(actor)).setColor(r, g, b);
			return true;
		}

		if (vst.sval.equals("size")) {
			// size takes one double argument	
			if (vst.nextToken() != TT_NUMBER) {
				System.out.println("Unexpected token for size at line "+
					st.lineno());
				return false;
			}
			
			((Polly)pollys.get(actor)).setSize(vst.nval);
			return true;
		}
	
		// Unkown keyword
		System.out.println("Unexpected keyword for actor at line "+st.lineno());
		return false;
	}


	private boolean readCamera(StreamTokenizer st) throws IOException {
		// read the next keyword
		if (st.nextToken() != TT_WORD) {
			System.out.println("Unexpected camera keyword "+st.toString()+
				" at line "+st.lineno());
			return false;
		}


		if (st.sval.equals("position")) {
			// FIXME: unimplemented for now 
			ParsedString value;
			value = resolveDefines(st);
			return true;
		}

		if (st.sval.equals("direction")) {
			// Read the parameters
			ParsedString value;
			value = resolveDefines(st);
			if (value.numtokens != 2) {
				System.out.println("Incorrect number of parameters for camera directions -- need 2: theta and phi. At line "+st.lineno());
				return false;
			}
			StreamTokenizer vst = new StreamTokenizer(new StringReader(value.str));
			boolean error = false;
			double t, p;
			if (vst.nextToken() != TT_NUMBER)
				error = true;
			t = vst.nval;
			if (vst.nextToken() != TT_NUMBER)
				error = true;
			p = vst.nval;
			
			if (error) {
				System.out.println("Unable to parse arguments for camera direction at line "+st.lineno());
				return false;
			}
	
			// Sets the globals to perform a rotation	
			theta = camtheta - t;
			phi = camphi - p;
			
			// Keep track of the new coordinates
			camtheta = t;
			camphi = p;
			return true;
		}

		System.out.println("Unknown camera direction keyword "+st.sval+
			 " at line "+st.lineno());
		return false;	
	}

	public void animate(double time) {
		// read the script to see if we need to do anything
		performActions(time);
		// animate all the pollys we have
		for (int i=0; i < pollys.size(); i++) 
			if (pollys.get(i) != null) 
				((Polly)pollys.get(i)).animate(time);
	}

	// This is the interface from javascript. We get a new script
	// to process here.
	public void writeScript(String a) {
		String s = unescape(a);
		recordString(s);
		st = new StreamTokenizer(new StringReader(s));
		super.stop();
		clearWorld();
		// ask for garbage collection of old objects
		System.runFinalization();
		System.gc();
		super.init();
		super.start();
	}

	// Does the reverse of the JavaScript escape() function.
	// This is so that we perserve linefeeds, etc. through the 
	// transmission.
	public String unescape(String s) {
		String ret = "";
		
		for (int i = 0; i < s.length(); i++) {
			// escape sequences look like %FF -- always 2 hex digits
			if (s.charAt(i) == '%') {
				i++;
				// maybe a %% -- escape the % ?
				if (s.charAt(i) == '%') {
					ret = ret + '%';
					continue;
				}
				ret = ret + (char)Integer.parseInt(s.substring(i,i+2), 16);
				i++;
			}
			else {   // otherwise treat it as a string, append to output
				ret = ret + s.charAt(i);
			}
		}		
		// Because of some weird StreamTokenizer bug, it needs a \n at the
		// end of the string to process lines correctly
		ret = ret + (char)0x0A;
		return ret;
	}

	// Records the string as a vector, so that we can print out line numbers
	private void recordString(String s) {
		int newline = -1, prev = 0;
		int numlines = 0;

		lines = new Vector();
		while ((newline = s.indexOf(0x0A, newline+1)) != -1) {
			numlines++;
			if (lines.size() < numlines)
				lines.setSize(numlines+1);
			lines.add(numlines, s.substring(prev, newline));
			prev = newline + 1;
		}
	}

	// Override the update method, to draw our own buttons.	
	public synchronized void update(Graphics g) {
		Rectangle r;

		super.update(g);
		for (int i = 0; i < buttons.size(); i++)
			if ((r = (Rectangle)buttons.get(i)) != null) {
				g.setColor(buttonColor);
				g.fill3DRect(r.x, r.y, r.width, r.height, selected==i);
				g.setColor(textColor);
				g.drawString("Actor "+i, r.x + 4, r.y + 15);
			}

		// draw the commands for the selected actor
		Vector linelist;
		if (lines != null && selected > 0 && selected < commands.size() &&
				(linelist = (Vector)commands.get(selected)) != null) {
			for (int i = 0; i < linelist.size(); i++) {
				g.setColor(i == 0 ? highlightColor : textColor);
				g.drawString((String)(lines.get(((Integer)linelist.get(i)).intValue())), 
					5, (170 - (i*15)));
			}
		}
    }

	private void brighten(int selected) {
		if (pollys.get(selected) != null) {
					float c[] = ((Color)colors.get(selected)).getRGBColorComponents(null);
					// brighten it
					c[0] += .7;
					c[1] += .7;
					c[2] += .7;
					if (c[0] > 1) c[0] = 1;
					if (c[1] > 1) c[1] = 1;
					if (c[2] > 1) c[2] = 1;
					((Polly)pollys.get(selected)).setColor(c[0], c[1], c[2]);
				}
	}

	private void resetColor(int selected) {
		if (pollys.get(selected) != null) {
					float c[] = ((Color)colors.get(selected)).getRGBColorComponents(null);
					((Polly)pollys.get(selected)).setColor(c[0], c[1], c[2]);
				}
	}

	// Override mouse actions to just handle our buttons.
	// We don't want the user to move the camera anyway.
	public boolean mouseDown(Event event, int x, int y) {
		if (y > (H - buttonHeight) && y < H) {
			int newsel = x / buttonWidth;
			if (newsel == selected) {
				resetColor(selected);
				selected = -1;
			}
			else {
				if (selected != -1)
					resetColor(selected);
				selected = newsel;
				brighten(selected);
			}
		}
		return true;
	}

	public boolean mouseUp(Event event, int x, int y) { return true; }
}
